from typing import Union

from authlib.integrations.starlette_client import OAuth
from fastapi import Request
from fastapi.security import HTTPBearer
from httpx import HTTPError
from pydantic import ValidationError
from tasteful.authn.base import AsyncAuthenticationBackend

from .user import OIDCUser


class OIDCAuthenticationBackend(AsyncAuthenticationBackend, HTTPBearer):
    metatada: dict

    def __init__(
        self,
        name: str = "oidc_app",
        metadata_url: str = "",
        client_id: str = "",
        client_secret: str = "",
        scopes: str = "",
        introspection_endpoint: str = "",
    ):
        super().__init__()

        self.name = name
        self.metadata_url = metadata_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.scopes = scopes
        self.introspection_endpoint = introspection_endpoint

        self.oauth = OAuth()
        self.oauth.register(
            self.name,
            client_id=self.client_id,
            client_secret=self.client_secret,
            server_metadata_url=self.metadata_url,
            introspection_endpoint=self.introspection_endpoint,
            client_kwargs={"scope": self.scopes},
        )

        self.client = self.oauth.create_client(self.name)

    async def authenticate(self, request: Request) -> Union[OIDCUser, None]:
        """Authenticate the user depending on the token in the request headers."""
        authorization = request.headers.get("Authorization")

        if not authorization:
            return None

        else:
            token = authorization.split(" ")[1]

            # This method calls the OIDC Backend (zitadel for example)
            # to authenticate the user
            user = await self._validate_token((token))
            return user
            
    async def _validate_token(self, token: str) -> Union[OIDCUser, None]:
        """Return the user info from the backend API."""
        try:
            if not self.introspection_endpoint:
                metadata = await self.client.load_server_metadata()
                self.introspection_endpoint = metadata.get("introspection_endpoint", "")

            result = (
                await self.client._get_oauth_client().introspect_token(
                    url=self.introspection_endpoint,
                    token=token,
                )
            ).json()

            if result.get("active", False):
                try:
                    return OIDCUser(user_info=result, name=result.get("name"))
                except ValidationError as e:
                    print(f"Validation Error: {e}")

        except HTTPError as e:
            print("Error happened when trying to call OIDC apps", e)
            return None

        return None



from typing import List, Type

from dependency_injector import providers
from fastapi import FastAPI, Security

from tasteful.base_flavor import BaseFlavor
from tasteful.config import TastefulConfig
from tasteful.containers.tasteful_container import TastefulContainer


class TastefulApp:
    def __init__(
        self,
        name: str = None,
        version: str = None,
        log_level: str = None,
        flavors: List[Type[BaseFlavor]] = [],
        authentication_backends: list[Security] = [],
    ):
        self.container = TastefulContainer()

        self.flavors = flavors
        self._register_flavors()

        self.container.config.from_pydantic(TastefulConfig())

        self.app = FastAPI(
            title=name or self.container.config.project.name() or "Tasteful Application",
            version=version or self.container.config.project.version() or "0.1.0",
            log_level=log_level or self.container.config.project.log_level() or "INFO",
            dependencies=authentication_backends,
        )

        self._inject_flavors()

    def _register_flavors(self) -> None:
        """Register all flavors with the app."""
        for flavor_class in self.flavors:
            setattr(
                self.container, flavor_class.__name__, providers.Singleton(flavor_class)
            )

    def _inject_flavors(self) -> None:
        """Inject flavors routes into the FastAPI application."""
        for flavor in self.flavors:
            instance = getattr(self.container, flavor.__name__)()
            self.app.include_router(instance.controller.router)



from typing import Literal, Type

from pydantic import BaseModel
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)


class ProjectSchema(BaseModel):
    name: str
    version: str
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class TastefulConfig(BaseSettings):
    model_config = SettingsConfigDict(
        extra="ignore",
        env_file=".env",
        env_file_encoding="utf-8",
        case_sensitive=False,
        env_nested_delimiter="__",
    )

    project: ProjectSchema

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        """Order the settings sources - only .env file."""
        return (
            init_settings,
            dotenv_settings,
        )



from pydantic_settings import (
    BaseSettings,
    SettingsConfigDict,
)


class BaseConfig(BaseSettings):
    model_config = SettingsConfigDict(
        extra="ignore",
        env_file=".env",
        env_file_encoding="utf-8",
        env_nested_delimiter="__",
        case_sensitive=False,
    )



import inspect

from typing import Type

from dependency_injector import providers

from tasteful.base_flavor.base_controller import BaseController
from tasteful.containers.base_flavor_container import BaseFlavorContainer
from tasteful.dependencies.graph import Graph
from tasteful.exceptions.injector import NotAvailableForInjectionExceptionError
from tasteful.repositories import BaseRepository

from .base_config import BaseConfig
from .base_service import BaseService


class BaseFlavor:
    """Base class for all Tasteful application flavors.

    Provides core functionality for defining specialized application components
    with their own configurations, service dependencies, and routing logic.
    """

    _injectable: list[Type]
    container: BaseFlavorContainer
    controller: Type[BaseController]

    def __init__(
        self,
        controller: Type[BaseController] = None,
        services: list[Type[BaseService]] = [],
        repositories: list[Type[BaseRepository]] = [],
        config: Type[BaseConfig] = None,
    ):
        """Initialize a new flavor instance."""
        self._injectable = services + repositories

        self.container = BaseFlavorContainer()

        if config:
            self.config = self._create_config(config)

        _graph = Graph(injectable_classes=self._injectable)
        _graph.resolve_dependencies()
        _dependencies = _graph.get_dep_order()

        self._register_dependencies(_dependencies)

        self.controller = self._create_controller(controller)

    def _register_dependencies(self, dependencies: list[Type]) -> None:
        """Register every services within the flavor's container."""
        for dependency in dependencies:
            self._register_dependency(dependency=dependency)

    def _register_dependency(self, dependency: Type) -> None:
        """Create an appropriate provider for a service within the flavor's container."""
        if dependency not in self._injectable:
            raise NotAvailableForInjectionExceptionError(
                f"Service {dependency} is not registered in the list of "
                f"injectable service of flavor {self.__class__.__name__}."
            )

        constructor_args = self._resolve_constructor_dependencies(dependency)

        setattr(
            self.container,
            dependency.__name__,
            providers.Singleton(dependency, **constructor_args),
        )

    def _resolve_constructor_dependencies(self, target_class: Type) -> dict:
        """Extract and resolve constructor dependencies for a class."""
        constructor_params = list(
            inspect.signature(target_class.__init__).parameters.values()
        )[1:]  # Exclude 'self' parameter

        # Filter out variadic parameters (*args, **kwargs)
        constructor_params = [
            param for param in constructor_params if param.name not in ["args", "kwargs"]
        ]

        dependency_providers = {}

        for param in constructor_params:
            dependency_type = param.annotation
            # Only process class types that are in our
            # injectable dependencies and exist in container

            if (
                dependency_type != inspect.Parameter.empty
                and inspect.isclass(dependency_type)
                and dependency_type in self._injectable
                and hasattr(self.container, dependency_type.__name__)
            ):
                dependency_providers[param.name] = getattr(
                    self.container, dependency_type.__name__
                ).provided

        return dependency_providers

    def _create_controller(
        self, controller: Type[BaseController]
    ) -> Type[BaseController]:
        dependency_providers = self._resolve_constructor_dependencies(controller)

        setattr(
            self.container,
            controller.__name__,
            providers.Singleton(controller, **dependency_providers),
        )

        return getattr(self.container, controller.__name__)()

    def _create_config(self, config: Type[BaseConfig]) -> BaseConfig:
        """Create a configuration instance for the flavor."""
        if not issubclass(config, BaseConfig):
            raise TypeError(f"{config} is not a subclass of BaseConfig")

        config_instance = config()

        setattr(self.container, config.__name__, providers.Configuration())
        getattr(self.container, config.__name__).from_pydantic(config_instance)

        self._injectable.append(config)

        return config_instance



from typing import Any, Callable


class EndpointDefinition:
    def __init__(self, method: str = "GET", *args: Any, **kwargs: Any):
        self.method = method
        self.path = args[0] if len(args) > 0 else kwargs.get("path")
        self.args = args[1:]
        self.kwargs = kwargs


class APIControllerDecorator:
    """Emulates FastAPI endpoint decorators for a controller class."""

    def __init__(self, method: str, *args: Any, **kwargs: Any) -> None:
        self.args = args
        self.kwargs = kwargs
        self.method = method

    def __call__(self, func: Callable[..., Any]) -> Any:
        """Call the decoration logic when class is placed as a decorators.

        Wraps the decorated function and add hidden `__endpoint_definitions__`
        attributes that contains all the args from the decorator.
        """
        allowed_methods = {
            "delete",
            "get",
            "head",
            "options",
            "patch",
            "post",
            "put",
            "trace",
        }
        if self.method not in allowed_methods:
            return NotImplementedError(
                f"Method {self.method} not allowed in APIController"
            )

        return self._intercept_method(func)

    def _intercept_method(self, endpoint: Callable[..., Any]) -> Callable[..., Any]:
        if not hasattr(endpoint, "__endpoint_definitions__"):
            endpoint.__endpoint_definitions__ = []
        endpoint.__endpoint_definitions__.append(
            EndpointDefinition(self.method, *self.args, **self.kwargs)
        )
        return endpoint


class Get(APIControllerDecorator):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("get", *args, **kwargs)


class Post(APIControllerDecorator):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("post", *args, **kwargs)


class Patch(APIControllerDecorator):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("patch", *args, **kwargs)


class Delete(APIControllerDecorator):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("delete", *args, **kwargs)


class Head(APIControllerDecorator):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("head", *args, **kwargs)


class Options(APIControllerDecorator):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("options", *args, **kwargs)


class Put(APIControllerDecorator):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("put", *args, **kwargs)


class Trace(APIControllerDecorator):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("trace", *args, **kwargs)



from dependency_injector import containers, providers


class TastefulContainer(containers.DynamicContainer):
    """Dependency injection container for Tasteful application."""

    config = providers.Configuration()
